from unittest import TestCase
from unittest.mock import Mock

import numpy as np
import pandas as pd

from rdt.transformers import BinaryEncoder


class TestBinaryEncoder(TestCase):

    def test___init__(self):
        """Test default instance"""
        # Run
        transformer = BinaryEncoder()

        # Asserts
        error_message = 'Unexpected missing_value_replacement'
        error_generation = 'Unexpected missing_value_generation'
        assert transformer.missing_value_replacement == 'mode', error_message
        assert transformer.missing_value_generation == 'random', error_generation

    def test___init___model_missing_value_passed(self):
        """Test when model missing value is passed to the init."""
        # Run
        transformer = BinaryEncoder(model_missing_values=True)

        # Assert
        transformer.missing_value_generation == 'from_column'

    def test__fit_missing_value_replacement_not_ignore(self):
        """Test _fit missing_value_replacement not equal to ignore"""
        # Setup
        data = pd.Series([False, True, True, False, True])

        # Run
        transformer = BinaryEncoder(missing_value_replacement=0)
        transformer._fit(data)

        # Asserts
        error_msg = 'Unexpected fill value'
        assert transformer.null_transformer._missing_value_replacement == 0, error_msg

    def test__fit_array(self):
        """Test _fit with numpy.array"""
        # Setup
        data = pd.Series([False, True, True, False, True])

        # Run
        transformer = BinaryEncoder(missing_value_replacement=0)
        transformer._fit(data)

        # Asserts
        error_msg = 'Unexpected fill value'
        assert transformer.null_transformer._missing_value_replacement == 0, error_msg

    def test__fit_missing_value_generation_from_column(self):
        """Test output_properties contains 'is_null' column.

        When missing_value_generation is 'from_column' the expected output is to have an extra
        column.
        """
        # Setup
        transformer = BinaryEncoder(missing_value_generation='from_column')
        data = pd.Series([True, np.nan])

        # Run
        transformer._fit(data)

        # Assert
        assert transformer.output_properties == {
            None: {'sdtype': 'float', 'next_transformer': None},
            'is_null': {'sdtype': 'float', 'next_transformer': None},
        }

    def test__transform_series(self):
        """Test transform pandas.Series"""
        # Setup
        data = pd.Series([False, True, None, True, False])

        # Run
        transformer = Mock()
        BinaryEncoder._transform(transformer, data)

        # Asserts
        expect_call_count = 1
        expect_call_args = pd.Series([0., 1., None, 1., 0.], dtype=float)

        error_msg = 'NullTransformer.transform must be called one time'
        assert transformer.null_transformer.transform.call_count == expect_call_count, error_msg
        pd.testing.assert_series_equal(
            transformer.null_transformer.transform.call_args[0][0],
            expect_call_args
        )

    def test__transform_array(self):
        """Test transform numpy.array"""
        # Setup
        data = pd.Series([False, True, None, True, False])

        # Run
        transformer = Mock()
        BinaryEncoder._transform(transformer, data)

        # Asserts
        expect_call_count = 1
        expect_call_args = pd.Series([0., 1., None, 1., 0.], dtype=float)

        error_msg = 'NullTransformer.transform must be called one time'
        assert transformer.null_transformer.transform.call_count == expect_call_count, error_msg
        pd.testing.assert_series_equal(
            transformer.null_transformer.transform.call_args[0][0],
            expect_call_args
        )

    def test__reverse_transform_missing_value_replacement_not_ignore(self):
        """Test _reverse_transform with missing_value_replacement not equal to ignore"""
        # Setup
        data = np.array([0.0, 1.0, 0.0, 1.0, 0.0])
        transformed_data = np.array([0.0, 1.0, 0.0, 1.0, 0.0])

        # Run
        transformer = Mock()
        transformer.missing_value_replacement = 0
        transformer.null_transformer.reverse_transform.return_value = transformed_data

        result = BinaryEncoder._reverse_transform(transformer, data)

        # Asserts
        expect = np.array([False, True, False, True, False])
        expect_call_count = 1

        np.testing.assert_equal(result, expect)

        error_msg = (
            'NullTransformer.reverse_transform should not be called when '
            'missing_value_replacement is ignore'
        )
        reverse_transform_call_count = transformer.null_transformer.reverse_transform.call_count
        assert reverse_transform_call_count == expect_call_count, error_msg

    def test__reverse_transform_series(self):
        """Test when data is a Series."""
        # Setup
        data = pd.Series([1., 0., 1.])

        # Run
        transformer = Mock()
        transformer.null_transformer.reverse_transform.return_value = data
        result = BinaryEncoder._reverse_transform(transformer, data)

        # Asserts
        expected = np.array([True, False, True])
        assert isinstance(result, pd.Series)
        np.testing.assert_equal(result.array, expected)

    def test__reverse_transform_not_null_values(self):
        """Test _reverse_transform not null values correctly"""
        # Setup
        data = np.array([1., 0., 1.])

        # Run
        transformer = Mock()
        transformer.null_transformer.reverse_transform.return_value = data

        result = BinaryEncoder._reverse_transform(transformer, data)

        # Asserts
        expected = np.array([True, False, True])

        assert isinstance(result, pd.Series)
        np.testing.assert_equal(result.array, expected)

    def test__reverse_transform_2d_ndarray(self):
        """Test _reverse_transform not null values correctly"""
        # Setup
        data = np.array([[1.], [0.], [1.]])

        # Run
        transformer = Mock()
        transformer.null_transformer.reverse_transform.return_value = data

        result = BinaryEncoder._reverse_transform(transformer, data)

        # Asserts
        expected = np.array([True, False, True])

        assert isinstance(result, pd.Series)
        np.testing.assert_equal(result.array, expected)

    def test__reverse_transform_float_values(self):
        """Test the ``_reverse_transform`` method with decimals.

        Expect that the ``_reverse_transform`` method handles decimal inputs
        correctly by rounding them.

        Input:
            - Transformed data with decimal values.
        Output:
            - Reversed transformed data.
        """
        # Setup
        data = np.array([1.2, 0.32, 1.01])
        transformer = Mock()
        transformer.null_transformer.reverse_transform.return_value = data

        # Run
        result = BinaryEncoder._reverse_transform(transformer, data)

        # Asserts
        expected = np.array([True, False, True])

        assert isinstance(result, pd.Series)
        np.testing.assert_equal(result.to_numpy(), expected)

    def test__reverse_transform_float_values_out_of_range(self):
        """Test the ``_reverse_transform`` method with decimals that are out of range.

        Expect that the ``_reverse_transform`` method handles decimal inputs
        correctly by rounding them. If the rounded decimal inputs are < 0 or > 1, expect
        expect them to be clipped.

        Input:
            - Transformed data with decimal values, some of which round to < 0 or > 1.
        Output:
            - Reversed transformed data.
        """
        # Setup
        data = np.array([1.9, -0.7, 1.01])
        transformer = Mock()
        transformer.null_transformer.reverse_transform.return_value = data

        # Run
        result = BinaryEncoder._reverse_transform(transformer, data)

        # Asserts
        expected = np.array([True, False, True])

        assert isinstance(result, pd.Series)
        np.testing.assert_equal(result.array, expected)

    def test__reverse_transform_numpy_nan(self):
        """Test the ``_reverse_transform`` method with decimals that are out of range.

        Expect that the ``_reverse_transform`` method contains the `np.nan` instead of
        other `nan` value.

        Input:
            - Transformed data with decimal values, some of which are ``np.nan``.

        Mock:
            - Mock `np.nan`.

        Output:
            - Reversed transformed data containing `np.nan` mocked value.
        """
        # Setup
        data = np.array([1.9, np.nan, 1.01])
        transformer = Mock()
        transformer.null_transformer.reverse_transform.return_value = data

        # Run
        result = BinaryEncoder._reverse_transform(transformer, data)

        # Asserts
        assert np.isnan(result[1])
        assert isinstance(result[1], float)



from rdt.transformers.pii.utils import get_provider_name, is_faker_function


def test_get_provider_name():
    """Test the ``get_provider_name`` method.

    Test that the function returns an expected provider name from the ``faker.Faker`` instance.
    If this is from the ``BaseProvider`` it should also return that name.
    """
    # Run
    email_provider = get_provider_name('email')
    lexify_provider = get_provider_name('lexify')

    # Assert
    assert email_provider == 'internet'
    assert lexify_provider == 'BaseProvider'


def test_is_faker_function():
    """Test that the method returns True if the ``function_name`` is a valid faker function."""
    # Run
    result = is_faker_function('address')

    # Assert
    assert result is True


def test_is_faker_function_error():
    """Test that the method returns False if ``function_name`` is not a valid faker function."""
    # Run
    result = is_faker_function('blah')

    # Assert
    assert result is False



from faker import Faker

from rdt.transformers.pii.anonymization import is_faker_function


def test_is_faker_function():
    """Test is_faker_function checks if function is a valid Faker function."""
    # Run
    result = is_faker_function('address')

    # Assert
    assert result is True


def test_is_faker_function_non_default_locale():
    """Test is_faker_function checks non-default locales."""
    # Setup
    function_name = 'postcode_in_province'

    # Run
    result = is_faker_function(function_name)

    # Assert
    assert result is True
    assert not hasattr(Faker(), function_name)



"""Basic Integration tests for the BaseTransformer."""

import numpy as np
import pandas as pd

from rdt.transformers.base import BaseMultiColumnTransformer, BaseTransformer


def test_dummy_transformer_series_output():
    """Test a transformer that outputs a Series.

    This validates that a Transformer that is implemented to
    produce a Series object as the output of the `_transform`
    method works as expected.

    Setup:
        - A DummyTransformer that implements a _transform
          that produces a Series and a reverse_transform
          that takes a Series as input.

    Input:
        - A DataFrame with a single column of boolean values

    Expected behavior:
        - The data should be transformed into a DataFrame that contains
          a single float column made of 0.0s and 1.0s.
        - The transformed data should be able to reversed to
          re-produce the input data.
    """
    # Setup
    class DummyTransformer(BaseTransformer):

        INPUT_SDTYPE = 'boolean'

        def _fit(self, data):
            pass

        def _transform(self, data):
            return data.astype(float)

        def _reverse_transform(self, data):
            return data.round() != 0

    # Run
    data = pd.DataFrame({
        'bool': [True, False, True, False]
    })

    transformer = DummyTransformer()
    transformed = transformer.fit_transform(data, 'bool')

    reverse = transformer.reverse_transform(transformed)

    # Assert
    expected_transform = pd.DataFrame({
        'bool': [1., 0., 1., 0.]
    })
    pd.testing.assert_frame_equal(expected_transform, transformed)
    pd.testing.assert_frame_equal(reverse, data)


def test_dummy_transformer_dataframe_output():
    """Test a transformer that outputs a DataFrame.

    This validates that a Transformer that is implemented to
    produce a DataFrame object as the output of the `_transform`
    method works as expected.

    Setup:
        - A DummyTransformer that implements a _transform
          that produces a DataFrame and a reverse_transform
          that takes a DataFrame as input.

    Input:
        - A DataFrame with a single column of boolean values

    Expected behavior:
        - The data should be transformed into a DataFrame that contains
          a float column made of 0.0s, 1.0s and -1.0s, and another float
          column that contains 0.0s and 1.0s indicating which values were
          null in the input.
        - The transformed data should be able to reversed to
          re-produce the input data.
    """
    # Setup
    class DummyTransformer(BaseTransformer):

        INPUT_SDTYPE = 'boolean'

        def __init__(self):
            super().__init__()
            self.output_properties = {
                None: {'sdtype': 'float'},
                'null': {'sdtype': 'float'},
            }

        def _fit(self, data):
            pass

        def _transform(self, data):
            out = pd.DataFrame(dict(zip(
                self.output_columns,
                [
                    data.astype(float).fillna(-1),
                    data.isna().astype(float)
                ]
            )))

            return out

        def _reverse_transform(self, data):
            output = data[self.output_columns[0]]
            output = output.round().astype(bool).astype(object)
            output.iloc[data[self.output_columns[1]] == 1] = np.nan

            return output

    data = pd.DataFrame({'bool': [True, False, True, np.nan]})
    transformer = DummyTransformer()

    # Run
    transformed = transformer.fit_transform(data, 'bool')
    reverse = transformer.reverse_transform(transformed)

    # Assert
    expected_transform = pd.DataFrame({
        'bool': [1., 0., 1., -1.],
        'bool.null': [0., 0., 0., 1.]
    })
    pd.testing.assert_frame_equal(expected_transform, transformed)
    pd.testing.assert_frame_equal(reverse, data)


def test_multi_column_transformer_same_number_of_columns_input_output():
    """Test a multi-column transformer when the same of input and output columns."""
    # Setup
    class AdditionTransformer(BaseMultiColumnTransformer):
        """This transformer takes 3 columns and return the cumulative sum of each row."""
        def _fit(self, columns_data):
            self.output_properties = {
                f'{self.columns[0]}': {'sdtype': 'numerical'},
                f'{self.columns[0]}+{self.columns[1]}': {'sdtype': 'numerical'},
                f'{self.columns[0]}+{self.columns[1]}+{self.columns[2]}': {'sdtype': 'numerical'}
            }

        def _get_prefix(self):
            return None

        def _transform(self, data):
            return data.cumsum(axis=1)

        def _reverse_transform(self, data):
            result = data.diff(axis=1)
            result.iloc[:, 0] = data.iloc[:, 0]

            return result.astype('int64')

    data_test = pd.DataFrame({
        'col_1': [1, 2, 3],
        'col_2': [10, 20, 30],
        'col_3': [100, 200, 300]
    })

    columns_to_sdtypes = {
        'col_1': 'numerical',
        'col_2': 'numerical',
        'col_3': 'numerical'
    }
    transformer = AdditionTransformer()

    # Run
    transformed = transformer.fit_transform(data_test, columns_to_sdtypes)
    reverse = transformer.reverse_transform(transformed)

    # Assert
    expected_transform = pd.DataFrame({
        'col_1': [1, 2, 3],
        'col_1+col_2': [11, 22, 33],
        'col_1+col_2+col_3': [111, 222, 333]
    })
    pd.testing.assert_frame_equal(expected_transform, transformed)
    pd.testing.assert_frame_equal(reverse, data_test)


def test_multi_column_transformer_less_output_than_input_columns():
    """Test a multi-column transformer when the output has less columns than the input."""
    class ConcatenateTransformer(BaseMultiColumnTransformer):
        """This transformer takes 4 columns and concatenate them into 2 columns.
        The two first and last columns are concatenated together.
        """
        def _fit(self, columns_data):
            self.name_1 = self.columns[0] + '#' + self.columns[1]
            self.name_2 = self.columns[2] + '#' + self.columns[3]
            self.output_properties = {
                f'{self.name_1}.concatenate_1': {'sdtype': 'categorical'},
                f'{self.name_2}.concatenate_2': {'sdtype': 'categorical'}
            }

        def _get_prefix(self):
            return None

        def _transform(self, data):
            data[self.name_1] = data.iloc[:, 0] + '#' + data.iloc[:, 1]
            data[self.name_2] = data.iloc[:, 2] + '#' + data.iloc[:, 3]

            return data.drop(columns=self.columns)

        def _reverse_transform(self, data):
            result = data.copy()
            column_names = list(data.columns)

            col1, col2 = column_names[0].split('#')
            result[[col1, col2]] = result[column_names[0]].str.split('#', expand=True)

            col3, col4 = column_names[1].split('#')
            result[[col3, col4]] = result[column_names[1]].str.split('#', expand=True)

            return result.drop(columns=column_names)

    data_test = pd.DataFrame({
        'col_1': ['A', 'B', 'C'],
        'col_2': ['D', 'E', 'F'],
        'col_3': ['G', 'H', 'I'],
        'col_4': ['J', 'K', 'L']
    })

    columns_to_sdtypes = {
        'col_1': 'categorical',
        'col_2': 'categorical',
        'col_3': 'categorical',
        'col_4': 'categorical'
    }
    transformer = ConcatenateTransformer()

    # Run
    transformer.fit(data_test, columns_to_sdtypes)
    transformed = transformer.transform(data_test)
    reverse = transformer.reverse_transform(transformed)

    # Assert
    expected_transform = pd.DataFrame({
        'col_1#col_2.concatenate_1': ['A#D', 'B#E', 'C#F'],
        'col_3#col_4.concatenate_2': ['G#J', 'H#K', 'I#L']
    })
    pd.testing.assert_frame_equal(expected_transform, transformed)
    pd.testing.assert_frame_equal(reverse, data_test)


def test_multi_column_transformer_more_output_than_input_columns():
    """Test a multi-column transformer when the output has more columns than the input."""
    class ExpandTransformer(BaseMultiColumnTransformer):

        def _fit(self, columns_data):
            self.output_properties = {
                f'{self.columns[0]}.first_part_1': {'sdtype': 'categorical'},
                f'{self.columns[0]}.second_part_1': {'sdtype': 'categorical'},
                f'{self.columns[1]}.first_part_2': {'sdtype': 'categorical'},
                f'{self.columns[1]}.second_part_2': {'sdtype': 'categorical'}
            }

        def _get_prefix(self):
            return None

        def _transform(self, data):
            data[self.output_columns[0]] = data[self.columns[0]].str[0]
            data[self.output_columns[1]] = data[self.columns[0]].str[1]
            data[self.output_columns[2]] = data[self.columns[1]].str[0]
            data[self.output_columns[3]] = data[self.columns[1]].str[1]

            return data.drop(columns=self.columns)

        def _reverse_transform(self, data):
            result = data.copy()
            reverse_1 = result[self.output_columns[0]] + result[self.output_columns[1]]
            reverse_2 = result[self.output_columns[2]] + result[self.output_columns[3]]
            result[self.columns[0]] = reverse_1
            result[self.columns[1]] = reverse_2

            return result.drop(columns=self.output_columns)

    data_test = pd.DataFrame({
        'col_1': ['AB', 'CD', 'EF'],
        'col_2': ['GH', 'IJ', 'KL'],
    })

    columns_to_sdtypes = {
        'col_1': 'categorical',
        'col_2': 'categorical'
    }
    transformer = ExpandTransformer()

    # Run
    transformer.fit(data_test, columns_to_sdtypes)
    transformed = transformer.transform(data_test)
    reverse = transformer.reverse_transform(transformed)

    # Assert
    expected_transform = pd.DataFrame({
        'col_1.first_part_1': ['A', 'C', 'E'],
        'col_1.second_part_1': ['B', 'D', 'F'],
        'col_2.first_part_2': ['G', 'I', 'K'],
        'col_2.second_part_2': ['H', 'J', 'L']
    })
    pd.testing.assert_frame_equal(expected_transform, transformed)
    pd.testing.assert_frame_equal(reverse, data_test)



import numpy as np
import pandas as pd

from rdt.transformers.datetime import OptimizedTimestampEncoder, UnixTimestampEncoder


class TestUnixTimestampEncoder:
    def test_unixtimestampencoder(self):
        """Test the ``UnixTimestampEncoder`` end to end."""
        # Setup
        ute = UnixTimestampEncoder(missing_value_replacement='mean')
        data = pd.DataFrame({'column': pd.to_datetime([None, '1996-10-17', '1965-05-23'])})

        # Run
        ute.fit(data, column='column')
        ute.set_random_state(np.random.RandomState(7), 'reverse_transform')
        transformed = ute.transform(data)
        reverted = ute.reverse_transform(transformed)

        # Asserts
        expected_transformed = pd.DataFrame({
            'column': [3.500064e+17, 845510400000000000, -145497600000000000]
        })

        pd.testing.assert_frame_equal(expected_transformed, transformed)
        pd.testing.assert_frame_equal(reverted, data)

    def test_unixtimestampencoder_different_format(self):
        """Test the ``UnixTimestampEncoder`` with a unique datetime format."""
        ute = UnixTimestampEncoder(missing_value_replacement='mean', datetime_format='%b %d, %Y')
        data = pd.DataFrame({'column': [None, 'Oct 17, 1996', 'May 23, 1965']})

        # Run
        ute.fit(data, column='column')
        ute.set_random_state(np.random.RandomState(7), 'reverse_transform')
        transformed = ute.transform(data)
        reverted = ute.reverse_transform(transformed)

        # Asserts
        expect_transformed = pd.DataFrame({
            'column': [3.500064e+17, 845510400000000000, -145497600000000000]
        })
        pd.testing.assert_frame_equal(expect_transformed, transformed)
        pd.testing.assert_frame_equal(reverted, data)

    def test_unixtimestampencoder_with_missing_value_generation_none(self):
        """Test that transformed data will replace nans with the mean."""
        # Setup
        ute = UnixTimestampEncoder(
            missing_value_replacement='mean',
            missing_value_generation=None,
            datetime_format='%b %d, %Y'
        )
        data = pd.DataFrame({'column': [None, 'Oct 17, 1996', 'May 23, 1965']})

        # Run
        ute.fit(data, column='column')
        ute.set_random_state(np.random.RandomState(7), 'reverse_transform')
        transformed = ute.transform(data)
        reverted = ute.reverse_transform(transformed)

        # Asserts
        expect_transformed = pd.DataFrame({
            'column': [3.500064e+17, 845510400000000000, -145497600000000000]
        })
        expected_reversed = pd.DataFrame({
            'column': ['Feb 03, 1981', 'Oct 17, 1996', 'May 23, 1965']
        })
        pd.testing.assert_frame_equal(expect_transformed, transformed)
        pd.testing.assert_frame_equal(reverted, expected_reversed)

    def test_unixtimestampencoder_with_missing_value_replacement_random(self):
        """Test that transformed data will replace nans with random values from the data."""
        # Setup
        ute = UnixTimestampEncoder(
            missing_value_replacement='random',
            datetime_format='%b %d, %Y'
        )
        data = pd.DataFrame({'column': [None, 'Oct 17, 1996', 'May 23, 1965']})

        # Run
        ute.fit(data, column='column')
        ute.set_random_state(np.random.RandomState(7), 'reverse_transform')
        transformed = ute.transform(data)
        reverted = ute.reverse_transform(transformed)

        # Asserts
        expect_transformed = pd.DataFrame({
            'column': [-7.007396e+16, 845510400000000000, -145497600000000000]
        })
        expected_reversed = pd.DataFrame({
            'column': [np.nan, 'Oct 17, 1996', 'May 23, 1965']
        })
        pd.testing.assert_frame_equal(expect_transformed, transformed)
        pd.testing.assert_frame_equal(reverted, expected_reversed)

    def test_unixtimestampencoder_with_model_missing_values(self):
        """Test that `model_missing_values` is accepted by the transformer."""
        # Setup
        ute = UnixTimestampEncoder('mean', True)
        data = pd.DataFrame({'column': pd.to_datetime([None, '1996-10-17', '1965-05-23'])})

        # Run
        ute.fit(data, column='column')
        ute.set_random_state(np.random.RandomState(7), 'reverse_transform')
        transformed = ute.transform(data)
        reverted = ute.reverse_transform(transformed)

        # Asserts
        expected_transformed = pd.DataFrame({
            'column': [3.500064e+17, 845510400000000000, -145497600000000000],
            'column.is_null': [1., 0., 0.]
        })

        pd.testing.assert_frame_equal(expected_transformed, transformed)
        pd.testing.assert_frame_equal(reverted, data)

    def test_unixtimestampencoder_with_integer_datetimes(self):
        """Test that the transformer properly handles integer columns."""
        # Setup
        ute = UnixTimestampEncoder('mean', True, datetime_format='%m%d%Y')
        data = pd.DataFrame({'column': [1201992, 11022028, 10011990]})

        # Run
        ute.fit(data, column='column')
        ute.set_random_state(np.random.RandomState(7), 'reverse_transform')
        transformed = ute.transform(data)
        reverted = ute.reverse_transform(transformed)

        # Asserts
        expected_transformed = pd.DataFrame({
            'column': [6.958656e+17, 1.856736e+18, 6.547392e+17],
        })

        pd.testing.assert_frame_equal(expected_transformed, transformed)
        pd.testing.assert_frame_equal(reverted, data)

    def test_unixtimestampencoder_with_nans(self):
        """Test that the transformer properly handles null columns."""
        # Setup
        ute = UnixTimestampEncoder('mean', True)
        data = pd.DataFrame({'column': [np.nan, np.nan, np.nan]})

        # Run
        ute.fit(data, column='column')
        ute.set_random_state(np.random.RandomState(7), 'reverse_transform')
        transformed = ute.transform(data)
        reverted = ute.reverse_transform(transformed)

        # Asserts
        expected_transformed = pd.DataFrame({
            'column': [0., 0., 0.],
            'column.is_null': [1., 1., 1.]
        })

        pd.testing.assert_frame_equal(expected_transformed, transformed)
        pd.testing.assert_frame_equal(reverted, data)

    def test_with_enforce_min_max_values_true(self):
        """Test that the transformer properly clipped out of bounds values."""
        # Setup
        ute = UnixTimestampEncoder(enforce_min_max_values=True)
        data = pd.DataFrame({'column': ['Feb 03, 1981', 'Oct 17, 1996', 'May 23, 1965']})
        ute.fit(data, column='column')

        # Run
        transformed = ute.transform(data)
        min_val = transformed['column'].min()
        max_val = transformed['column'].max()
        transformed.loc[transformed['column'] == min_val, 'column'] = min_val - 1e17
        transformed.loc[transformed['column'] == max_val, 'column'] = max_val + 1e17
        reverted = ute.reverse_transform(transformed)

        # Asserts
        assert ute._min_value == min_val
        assert ute._max_value == max_val
        pd.testing.assert_frame_equal(reverted, data)


class TestOptimizedTimestampEncoder:
    def test_optimizedtimestampencoder(self):
        ote = OptimizedTimestampEncoder(missing_value_replacement='mean')
        data = pd.DataFrame({'column': pd.to_datetime([None, '1996-10-17', '1965-05-23'])})

        # Run
        ote.fit(data, column='column')
        ote.set_random_state(np.random.RandomState(7), 'reverse_transform')
        transformed = ote.transform(data)
        reverted = ote.reverse_transform(transformed)

        # Asserts
        expect_transformed = pd.DataFrame({'column': [4051.0, 9786.0, -1684.0]})
        pd.testing.assert_frame_equal(expect_transformed, transformed)
        pd.testing.assert_frame_equal(reverted, data)



import platform
from unittest.mock import Mock, patch

import numpy as np
import pandas as pd
import pytest

from rdt.transformers.datetime import OptimizedTimestampEncoder, UnixTimestampEncoder
from rdt.transformers.null import NullTransformer


class TestUnixTimestampEncoder:

    def test___init__(self):
        """Test the ``__init__`` method and the passed arguments are stored as attributes."""
        # Run
        transformer = UnixTimestampEncoder(
            missing_value_replacement='mode',
            missing_value_generation='from_column',
            datetime_format='%M-%d-%Y',
            enforce_min_max_values=True,
        )

        # Asserts
        assert transformer.missing_value_replacement == 'mode'
        assert transformer.missing_value_generation == 'from_column'
        assert transformer.datetime_format == '%M-%d-%Y'
        assert transformer.enforce_min_max_values is True

    def test___init__with_model_missing_values(self):
        """Test the ``__init__`` method and the passed arguments are stored as attributes."""
        # Run
        transformer = UnixTimestampEncoder(
            missing_value_replacement='mode',
            model_missing_values=False,
            datetime_format='%M-%d-%Y'
        )

        # Asserts
        assert transformer.missing_value_replacement == 'mode'
        assert transformer.missing_value_generation == 'random'
        assert transformer.datetime_format == '%M-%d-%Y'

    def test__convert_to_datetime(self):
        """Test the ``_convert_to_datetime`` method.

        Test to make sure the transformer converts the data to datetime
        if it is of type ``object`` and can be converted.

        Input:
            - a pandas Series of dtype object, with elements that can be
            converted to datetime.

        Output:
            - a pandas series of type datetime.
        """
        # Setup
        data = pd.Series(['2020-01-01', '2020-02-01', '2020-03-01'])
        transformer = UnixTimestampEncoder()

        # Run
        converted_data = transformer._convert_to_datetime(data)

        # Assert
        expected_data = pd.Series(pd.to_datetime(['2020-01-01', '2020-02-01', '2020-03-01']))
        pd.testing.assert_series_equal(expected_data, converted_data)

    def test__convert_to_datetime_format(self):
        """Test the ``_convert_to_datetime`` method.

        Test to make sure the transformer uses the specified format
        to convert the data to datetime.

        Setup:
            - The transformer will be initialized with a format.
        Input:
            - a pandas Series of dtype object, with elements in the specified
            format.

        Output:
            - a pandas series of type datetime.
        """
        # Setup
        data = pd.Series(['01Feb2020', '02Mar2020', '03Jan2010'])
        dt_format = '%d%b%Y'
        transformer = UnixTimestampEncoder(datetime_format=dt_format)

        # Run
        converted_data = transformer._convert_to_datetime(data)

        # Assert
        expected_data = pd.Series(pd.to_datetime(['01Feb2020', '02Mar2020', '03Jan2010']))
        pd.testing.assert_series_equal(expected_data, converted_data)

    def test__convert_to_datetime_not_convertible_raises_error(self):
        """Test the ``_convert_to_datetime`` method.

        Test to make sure a ``TypeError`` is raised if the data is of type
        ``object`` but can't be converted.

        Input:
            - a pandas Series of dtype object, with elements that can't be
            converted to datetime.

        Expected behavior:
            - a ``TypeError`` is raised.
        """
        # Setup
        data = pd.Series(['2020-01-01-can', '2020-02-01-not', '2020-03-01-convert'])
        transformer = UnixTimestampEncoder()

        # Run
        error_message = 'Data must be of dtype datetime, or castable to datetime.'
        with pytest.raises(TypeError, match=error_message):
            transformer._convert_to_datetime(data)

    def test__convert_to_datetime_wrong_format_raises_error(self):
        """Test the ``_convert_to_datetime`` method.

        Test to make sure the transformer raises an error if the data does
        not match the specified format.

        Setup:
            - The transformer will be initialized with a format.
        Input:
            - a pandas Series of dtype object, with elements in the wrong
            format.

        Output:
            - a pandas series of type datetime.
        """
        # Setup
        data = pd.Series(['01-02-2020', '02-03-2020', '03J-01-2010'])
        dt_format = '%d%b%Y'
        transformer = UnixTimestampEncoder(datetime_format=dt_format)

        # Run
        error_message = 'Data does not match specified datetime format.'
        with pytest.raises(ValueError, match=error_message):
            transformer._convert_to_datetime(data)

    def test__transform_helper_calls_convert_to_datetime(self):
        """Test the ``_transform_helper`` method.

        Validate the helper transformer produces the correct value with ``strip_constant`` True.

        Input:
            - a pandas series of datetimes.

        Output:
            - a pandas series of the transformed datetimes.
        """
        # Setup
        data = pd.to_datetime(['2020-01-01', '2020-02-01', '2020-03-01'])
        transformer = UnixTimestampEncoder()
        transformer._convert_to_datetime = Mock()
        transformer._convert_to_datetime.return_value = data

        # Run
        transformer._transform_helper(data)

        # Assert
        transformer._convert_to_datetime.assert_called_once_with(data)

    def test__transform_helper(self):
        """Test the ``_transform_helper`` method.

        Validate the helper transformer produces the correct values.

        Input:
            - a pandas series of datetimes.

        Output:
            - a pandas series of the transformed datetimes.
        """
        # Setup
        data = pd.to_datetime(['2020-01-01', '2020-02-01', '2020-03-01'])
        transformer = UnixTimestampEncoder()

        # Run
        transformed = transformer._transform_helper(data)

        # Assert
        np.testing.assert_allclose(transformed, np.array([
            1.577837e+18, 1.580515e+18, 1.583021e+18,
        ]), rtol=1e-5)

    def test__reverse_transform_helper_nulls(self):
        """Test the ``_reverse_transform_helper`` with null values.

        Setup:
            - Mock the ``instance.null_transformer``.
            - Set the ``missing_value_replacement``.

        Input:
            - a pandas series.

        Output:
            - a pandas datetime index.

        Expected behavior:
            - The mock should call its ``reverse_transform`` method.
        """
        # Setup
        data = pd.to_datetime(['2020-01-01', '2020-02-01', '2020-03-01'])
        transformer = UnixTimestampEncoder(missing_value_replacement='mean')
        transformer.null_transformer = Mock()
        transformer.null_transformer.reverse_transform.return_value = pd.Series([1, 2, 3])

        # Run
        transformer._reverse_transform_helper(data)

        # Assert
        transformer.null_transformer.reverse_transform.assert_called_once()
        datetimes = transformer.null_transformer.reverse_transform.mock_calls[0][1][0]
        np.testing.assert_array_equal(data.to_numpy(), datetimes)

    def test__reverse_transform_helper_model_missing_values_true(self):
        """Test the ``_reverse_transform_helper`` with null values.

        Setup:
            - Mock the ``instance.null_transformer``.
            - Set the ``model_missing_values``.

        Input:
            - a pandas series.

        Output:
            - a pandas datetime index.

        Expected behavior:
            - The mock should call its ``reverse_transform`` method.
        """
        # Setup
        data = pd.to_datetime(['2020-01-01', '2020-02-01', '2020-03-01'])
        transformer = UnixTimestampEncoder(model_missing_values=True)
        transformer.null_transformer = Mock()
        transformer.null_transformer.reverse_transform.return_value = pd.Series([1, 2, 3])

        # Run
        transformer._reverse_transform_helper(data)

        # Assert
        transformer.null_transformer.reverse_transform.assert_called_once()
        datetimes = transformer.null_transformer.reverse_transform.mock_calls[0][1][0]
        np.testing.assert_array_equal(data.to_numpy(), datetimes)

    @patch('rdt.transformers.datetime.NullTransformer')
    def test__fit(self, null_transformer_mock):
        """Test the ``_fit`` method for numpy arrays.

        Validate that this method (1) sets ``self.null_transformer`` to the
        ``NullTransformer`` with the correct attributes, (2) calls the
        ``self.null_transformer``'s ``fit`` method, and (3) calls the
        ``_transform_helper`` method with the transformed data.
        """
        # Setup
        data = pd.to_datetime(['2020-01-01', '2020-02-01', '2020-03-01'])
        transformer = UnixTimestampEncoder()

        # Run
        transformer._fit(data)

        # Assert
        null_transformer_mock.assert_called_once_with('mean', 'random')
        assert null_transformer_mock.return_value.fit.call_count == 1
        np.testing.assert_allclose(
            null_transformer_mock.return_value.fit.call_args_list[0][0][0],
            np.array([1.577837e+18, 1.580515e+18, 1.583021e+18]), rtol=1e-5
        )

    def test__fit_enforce_min_max_values(self):
        """Test the ``_fit`` method when enforce_min_max_values is True.

        It should compute the min and max values of the integer conversion
        of the datetimes.
        """
        # Setup
        data = pd.to_datetime(['2020-01-01', '2020-02-01', '2020-03-01'])
        transformer = UnixTimestampEncoder(enforce_min_max_values=True)

        # Run
        transformer._fit(data)

        # Assert
        assert transformer._min_value == 1.5778368e+18
        assert transformer._max_value == 1.5830208e+18

    def test__fit_calls_transform_helper(self):
        """Test the ``_fit`` method.

        The ``_fit`` method should call the ``_transform_helper`` method.
        """
        # Setup
        data = pd.to_datetime(['2020-01-01', '2020-02-01', '2020-03-01'])
        transformer = UnixTimestampEncoder()
        transformer._transform_helper = Mock()
        transformer._transform_helper.return_value = pd.Series(data)

        # Run
        transformer._fit(data)

        # Assert
        transformer._transform_helper.assert_called_once()
        assert transformer.output_properties == {
            None: {'sdtype': 'float', 'next_transformer': None},
        }

    @patch('rdt.transformers.datetime._guess_datetime_format_for_array')
    def test__fit_calls_guess_datetime_format(self, mock__guess_datetime_format_for_array):
        """Test the ``_fit`` method.

        The ``_fit`` method should call the ``_transform_helper`` method.
        """
        # Setup
        data = pd.Series(['2020-02-01', '2020-03-01'])
        mock__guess_datetime_format_for_array.return_value = '%Y-%m-%d'
        transformer = UnixTimestampEncoder()

        # Run
        transformer._fit(data)

        # Assert
        np.testing.assert_array_equal(
            mock__guess_datetime_format_for_array.call_args[0][0],
            np.array(['2020-02-01', '2020-03-01'])
        )
        assert transformer.datetime_format == '%Y-%m-%d'

    def test__fit_missing_value_generation(self):
        """Test output_properties contains 'is_null' column.

        When missing_value_generation is 'from_column' the expected output is to have an extra
        column.
        """
        # Setup
        transformer = UnixTimestampEncoder(missing_value_generation='from_column')
        data = pd.Series(['2020-02-01', np.nan])

        # Run
        transformer._fit(data)

        # Assert
        assert transformer.output_properties == {
            None: {'sdtype': 'float', 'next_transformer': None},
            'is_null': {'sdtype': 'float', 'next_transformer': None},
        }

    def test__transform(self):
        """Test the ``_transform`` method for numpy arrays.

        Validate that this method calls the ``null_transformer.transform`` method.
        It should also check that the final output is correct.
        """
        # Setup
        data = pd.to_datetime(['2020-01-01', '2020-02-01', '2020-03-01'])
        transformer = UnixTimestampEncoder()
        transformer.null_transformer = Mock()

        # Run
        transformer._transform(data)

        # Assert
        assert transformer.null_transformer.transform.call_count == 1
        np.testing.assert_allclose(
            transformer.null_transformer.transform.call_args_list[0][0],
            np.array([[1.577837e+18, 1.580515e+18, 1.583021e+18]]), rtol=1e-5
        )

    def test__reverse_transform_all_none(self):
        """Test the ``_reverse_transform`` method with ``None`` values.

        Validate that the method transforms ``None`` into ``NaT``.
        """
        # Setup
        ute = UnixTimestampEncoder()
        ute.null_transformer = NullTransformer('mean')

        # Run
        output = ute._reverse_transform(pd.Series([None]))

        # Assert
        expected = pd.Series(pd.to_datetime(['NaT']))
        pd.testing.assert_series_equal(output, expected)

    def test__reverse_transform(self):
        """Test the ``_reverse_transform`` method.

        Validate that the method correctly reverse transforms.
        """
        # Setup
        ute = UnixTimestampEncoder()
        transformed = np.array([1.5778368e+18, 1.5805152e+18, 1.5830208e+18])
        ute.null_transformer = NullTransformer('mean')

        # Run
        output = ute._reverse_transform(transformed)

        # Assert
        expected = pd.Series(pd.to_datetime(['2020-01-01', '2020-02-01', '2020-03-01']))
        pd.testing.assert_series_equal(output, expected)

    def test__reverse_transform_enforce_min_max_values(self):
        """Test the ``_reverse_transform`` with enforce_min_max_values True.

        All the values that are outside the min and max values should be clipped to the min and
        max values.
        """
        # Setup
        ute = UnixTimestampEncoder(enforce_min_max_values=True)
        transformed = np.array([
            1.5678367e+18, 1.5778368e+18, 1.5805152e+18, 1.5830208e+18, 1.5930209e+18
        ])
        ute.null_transformer = NullTransformer('mean')
        ute._min_value = 1.5778368e+18
        ute._max_value = 1.5830208e+18

        # Run
        output = ute._reverse_transform(transformed)

        # Assert
        expected = pd.Series(pd.to_datetime([
            '2020-01-01', '2020-01-01', '2020-02-01', '2020-03-01', '2020-03-01'
        ]))
        pd.testing.assert_series_equal(output, expected)

    def test__reverse_transform_datetime_format_dtype_is_datetime(self):
        """Test the ``_reverse_transform`` method returns the correct datetime format."""
        # Setup
        ute = UnixTimestampEncoder()
        ute.datetime_format = '%b %d, %Y'
        transformed = np.array([1.5778368e+18, 1.5805152e+18, 1.5830208e+18])
        ute._dtype = np.dtype('<M8[ns]')
        ute.null_transformer = NullTransformer('mean')

        # Run
        output = ute._reverse_transform(transformed)

        # Assert
        expected = pd.Series(pd.to_datetime(['Jan 01, 2020', 'Feb 01, 2020', 'Mar 01, 2020']))
        pd.testing.assert_series_equal(output, expected)

    def test__reverse_transform_datetime_format(self):
        """Test the ``_reverse_transform`` method returns the correct datetime format."""
        # Setup
        ute = UnixTimestampEncoder()
        ute.datetime_format = '%b %d, %Y'
        transformed = np.array([1.5778368e+18, 1.5805152e+18, 1.5830208e+18])
        ute._dtype = 'object'
        ute.null_transformer = NullTransformer('mean')

        # Run
        output = ute._reverse_transform(transformed)

        # Assert
        expected = pd.Series(['Jan 01, 2020', 'Feb 01, 2020', 'Mar 01, 2020'])
        pd.testing.assert_series_equal(output, expected)

    def test__reverse_transform_datetime_format_with_strftime_formats(self):
        """Test the ``_reverse_transform`` method returns the correct datetime format."""
        # Setup
        ute = UnixTimestampEncoder()
        ute.datetime_format = '%b %-d, %Y'
        transformed = np.array([1.5778368e+18, 1.5805152e+18, 1.5830208e+18])
        ute._dtype = 'object'
        ute.null_transformer = NullTransformer('mean')

        # Run
        output = ute._reverse_transform(transformed)

        # Assert
        expected = pd.Series(['Jan 1, 2020', 'Feb 1, 2020', 'Mar 1, 2020'])
        if 'windows' not in platform.system().lower():
            pd.testing.assert_series_equal(output, expected)

    def test__reverse_transform_datetime_format_with_nans(self):
        """Test the ``_reverse_transform`` method returns the correct datetime format with nans."""
        # Setup
        ute = UnixTimestampEncoder()
        ute.datetime_format = '%b %d, %Y'
        transformed = np.array([1.5778368e+18, 1.5805152e+18, np.nan])
        ute._dtype = 'object'
        ute.null_transformer = NullTransformer('mean')

        # Run
        output = ute._reverse_transform(transformed)

        # Assert
        expected = pd.Series(['Jan 01, 2020', 'Feb 01, 2020', np.nan])
        pd.testing.assert_series_equal(output, expected)

    def test__reverse_transform_only_nans(self):
        """Test the ``_reverse_transform`` method returns the correct datetime format with nans."""
        # Setup
        ute = UnixTimestampEncoder()
        transformed = np.array([np.nan, np.nan, np.nan])
        ute._dtype = 'float'
        ute.null_transformer = NullTransformer('mean')

        # Run
        output = ute._reverse_transform(transformed)

        # Assert
        expected = pd.Series([np.nan, np.nan, np.nan])
        pd.testing.assert_series_equal(output, expected)


class TestOptimizedTimestampEncoder:

    def test___init__(self):
        """Test the ``__init__`` method."""
        # Run
        transformer = OptimizedTimestampEncoder(
            missing_value_replacement='mode',
            missing_value_generation='from_column',
            datetime_format='%M-%d-%Y',
            enforce_min_max_values=True,
        )

        # Asserts
        assert transformer.enforce_min_max_values is True
        assert transformer.missing_value_replacement == 'mode'
        assert transformer.missing_value_generation == 'from_column'
        assert transformer.datetime_format == '%M-%d-%Y'
        assert transformer.divider is None
        assert transformer.null_transformer is None

    def test__find_divider(self):
        """Test the ``_find_divider`` method.

        Find the greatest common denominator out of these values: [10] * 9 + [60, 60, 24],
        where each consecutive value in the list is multiplied by the previous one
        (so 10, 100, 1000, etc).
        """
        # Setup
        data = np.array([100, 7919])
        transformer = OptimizedTimestampEncoder()

        # Run
        transformer._find_divider(data)

        # Assert
        assert transformer.divider == 1

    def test__transform_helper(self):
        """Test the ``_transform_helper`` method.

        Validate the helper method produces the values stripped to the smallest
        non-zero time unit.
        """
        # Setup
        data = pd.to_datetime(['2020-01-01', '2020-02-01', '2020-03-01'])
        transformer = OptimizedTimestampEncoder()

        # Run
        transformed = transformer._transform_helper(data)

        # Assert
        np.testing.assert_allclose(transformed, np.array([
            18262., 18293., 18322.,
        ]))

    def test__reverse_transform_helper(self):
        """Test the ``_reverse_transform_helper`` method.

        Validate the helper produces the values multiplied by the
        smallest non-zero time unit.
        """
        # Setup
        data = pd.Series([18262., 18293., 18322.])
        transformer = OptimizedTimestampEncoder()
        transformer.divider = 1000
        transformer.null_transformer = Mock()
        transformer.null_transformer.reverse_transform.side_effect = lambda x: x

        # Run
        multiplied = transformer._reverse_transform_helper(data)

        # Assert
        expected = np.array([18262000, 18293000, 18322000])
        np.testing.assert_allclose(multiplied, expected)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

import json
import os
from glob import glob
from setuptools import setup, find_namespace_packages

with open('README.md', encoding='utf-8') as readme_file:
    readme = readme_file.read()

with open('HISTORY.md', encoding='utf-8') as history_file:
    history = history_file.read()

install_requires = [
    "numpy>=1.20.0,<2;python_version<'3.10'",
    "numpy>=1.23.3,<2;python_version>='3.10'",
    "pandas>=1.1.3;python_version<'3.10'",
    "pandas>=1.3.4;python_version>='3.10' and python_version<'3.11'",
    "pandas>=1.5.0;python_version>='3.11'",
    "scipy>=1.5.4,<2;python_version<'3.10'",
    "scipy>=1.9.2,<2;python_version>='3.10'",
    "scikit-learn>=0.24,<2;python_version<'3.10'",
    "scikit-learn>=1.1.3,<2;python_version>='3.10'",
    'Faker>=17,<20',
]

copulas_requires = [
    'copulas>=0.9.0,<0.10',
]

setup_requires = [
    'pytest-runner>=2.11.1',
]

tests_require = [
    'pytest>=3.4.2',
    'pytest-cov>=2.6.0',
    'jupyter>=1.0.0,<2',
    'rundoc>=0.4.3,<0.5',
    'pytest-subtests>=0.5,<1.0',
]

addons_require = []

for addon_json in glob('rdt/transformers/addons/*/*.json'):
    with open(addon_json, 'r', encoding='utf-8') as addon_json_file:
        requirements = json.load(addon_json_file).get('requirements')
        if requirements:
            addons_require.extend(requirements)

development_requires = [
    # general
    'bumpversion>=0.5.3,<0.6',
    'pip>=9.0.1',
    'watchdog>=0.8.3,<0.11',

    # style check
    'pycodestyle<2.8.0,>=2.7.0',
    'pyflakes<2.4.0,>=2.3.0',
    'flake8>=3.7.7,<4',
    'flake8-absolute-import>=1.0,<2',
    'flake8-builtins>=1.5.3,<1.6',
    'flake8-comprehensions>=3.6.1,<3.7',
    'flake8-debugger>=4.0.0,<4.1',
    'flake8-docstrings>=1.5.0,<2',
    'flake8-mock>=0.3,<0.4',
    'flake8-variables-names>=0.0.4,<0.1',
    'dlint>=0.11.0,<0.12',  # code security addon for flake8
    'flake8-fixme>=1.1.1,<1.2',
    'flake8-eradicate>=1.1.0,<1.2',
    'flake8-mutable>=1.2.0,<1.3',
    'flake8-print>=4.0.0,<4.1',
    'isort>=4.3.4,<5',
    'pylint>=2.5.3,<3',
    'pandas-vet>=0.2.3,<0.3',
    'flake8-multiline-containers>=0.0.18,<0.1',
    'flake8-pytest-style>=1.5.0,<2',
    'flake8-quotes>=3.3.0,<4',
    'flake8-expression-complexity>=0.0.9,<0.1',
    'pep8-naming>=0.12.1,<0.13',
    'pydocstyle>=6.1.1,<6.2',
    'flake8-sfs>=0.0.3,<0.1',

    # fix style issues
    'autoflake>=1.1,<2',
    'autopep8>=1.4.3,<1.6',

    # distribute on PyPI
    'twine>=1.10.0,<4',
    'wheel>=0.30.0',

    # Advanced testing
    'coverage>=4.5.1,<6',
    'tox>=2.9.1,<4',
    'tabulate>=0.8.9,<1',

    # Invoking test commands
    'invoke'
]

setup(
    author='DataCebo, Inc.',
    author_email='info@sdv.dev',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: Free for non-commercial use',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    description='Reversible Data Transforms',
    extras_require={
        'copulas': copulas_requires,
        'test': tests_require + copulas_requires + addons_require,
        'dev': development_requires + tests_require + copulas_requires + addons_require,
    },
    include_package_data=True,
    install_requires=install_requires,
    keywords='rdt',
    license='BSL-1.1',
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/markdown',
    name='rdt',
    packages=find_namespace_packages(
        include=['rdt', 'rdt.*'],
        exclude=['rdt.transformers.addons.*']
    ),
    python_requires='>=3.8,<3.12',
    setup_requires=setup_requires,
    test_suite='tests',
    tests_require=tests_require,
    url='https://github.com/sdv-dev/RDT',
    version='1.9.1.dev0',
    zip_safe=False,
)



from unittest.mock import Mock, patch

from rdt.transformers.pii.anonymization import (
    _detect_provider_name, get_anonymized_transformer, get_faker_instance, is_faker_function)


class TestAnonimization:

    def test__detect_provider_name(self):
        """Test the ``_detect_provider_name`` method.

        Test that the function returns an expected provider name from the ``faker.Faker`` instance.
        If this is from the ``BaseProvider`` it should also return that name.

        Input:
            - Faker function name.

        Output:
            - The faker provider name for that function.
        """
        # Run / Assert
        email_provider = _detect_provider_name('email')
        lexify_provider = _detect_provider_name('lexify')
        state_provider = _detect_provider_name('state')

        assert email_provider == 'internet'
        assert lexify_provider == 'BaseProvider'
        assert state_provider == 'address.en_US'

    @patch('rdt.transformers.pii.anonymization.AnonymizedFaker')
    def test_get_anonymized_transformer_with_existing_sdtype(self, mock_anonymized_faker):
        """Test the ``get_anonymized_transformer`` method.

        Test that when calling with an existing ``sdtype`` / ``function_name`` from the
        ``SDTYPE_ANONYMZIERS`` dictionary, their ``provider_name`` and ``function_name`` are being
        used by default, and also other ``kwargs`` and provided locales are being passed to the
        ``AnonymizedFaker``.

        Input:
            - ``function_name`` from the ``SDTYPE_ANONYMIZERS``.
            - ``function_kwargs`` additional keyword arguments for that set of arguments.

        Mock:
            - Mock ``AnonymizedFaker`` and assert that has been called with the expected
              arguments.

        Output:
            - The return value must be the instance of ``AnonymizedFaker``.
        """
        # Setup
        output = get_anonymized_transformer('email', transformer_kwargs={
            'function_kwargs': {'domain': '@gmail.com'}, 'locales': ['en_CA', 'fr_CA']
        })

        # Assert
        assert output == mock_anonymized_faker.return_value
        mock_anonymized_faker.assert_called_once_with(
            provider_name='internet',
            function_name='email',
            function_kwargs={'domain': '@gmail.com'},
            locales=['en_CA', 'fr_CA']
        )

    @patch('rdt.transformers.pii.anonymization.AnonymizedFaker')
    def test_get_anonymized_transformer_with_custom_sdtype(self, mock_anonymized_faker):
        """Test the ``get_anonymized_transformer`` method.

        Test that when calling with a custom ``sdtype`` / ``function_name`` that does not belong
        to the ``SDTYPE_ANONYMZIERS`` dictionary. The ``provider_name`` is being found
        automatically other ``kwargs`` and provided locales are being passed to the
        ``AnonymizedFaker``.

        Input:
            - ``function_name`` color.
            - ``function_kwargs`` a dictionary with ``'hue': 'red'``.

        Mock:
            - Mock ``AnonymizedFaker`` and assert that has been called with the expected
              arguments.

        Output:
            - The return value must be the instance of ``AnonymizedFaker``.
        """
        # Setup
        output = get_anonymized_transformer('color', transformer_kwargs={
            'function_kwargs': {'hue': 'red'}, 'locales': ['en_CA', 'fr_CA']
        })

        # Assert
        assert output == mock_anonymized_faker.return_value
        mock_anonymized_faker.assert_called_once_with(
            provider_name='color',
            function_name='color',
            function_kwargs={'hue': 'red'},
            locales=['en_CA', 'fr_CA']
        )

    @patch('rdt.transformers.pii.anonymization.Faker')
    def test_is_faker_function(self, faker_mock):
        """Test that the method returns True if the ``function_name`` is a valid faker function.

        This test mocks the ``Faker`` method to make sure that the ``function_name`` is an
        attribute it has.
        """
        # Setup
        faker_mock.return_value = Mock(spec=['address'])

        # Run
        result = is_faker_function('address')

        # Assert
        assert result is True

    @patch('rdt.transformers.pii.anonymization.get_faker_instance')
    def test_is_faker_function_error(self, mock_get_faker_instance):
        """Test that the method returns False if ``function_name`` is not a valid faker function.

        If the ``function_name`` is not an attribute of ``Faker()`` then we should return false.
        This test mocks ``Faker`` to not have the attribute that is passed as ``function_name``.
        """
        # Setup
        mock_get_faker_instance.return_value = Mock(spec=[])

        # Run
        result = is_faker_function('blah')

        # Assert
        assert result is False
        mock_get_faker_instance.assert_called_once()

    @patch('rdt.transformers.pii.anonymization.Faker')
    def test_get_faker_instance(self, mock_faker):
        """Test that ``get_faker_instance`` returns the same object."""
        # Setup
        first_instance = get_faker_instance()

        # Run
        second_instance = get_faker_instance()

        # Assert
        assert id(first_instance) == id(second_instance)
